#!/usr/bin/env python

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
    CLI interface for barbican management
"""
from __future__ import print_function

from os_vm_expire.common import config
# from os_vm_expire.model import clean
# from os_vm_expire.model.migration import commands
from os_vm_expire.model.models import VmExclude
from os_vm_expire.model import repositories
import os_vm_expire.version

import argparse
import datetime
import prettytable
import six
import sys
import time

from oslo_config import cfg
from oslo_db import options
from oslo_log import log as logging
from oslo_utils import encodeutils


CONF = cfg.CONF
options.set_defaults(CONF)
LOG = logging.getLogger(__name__)


# Decorators for actions
def args(*args, **kwargs):
    def _decorator(func):
        func.__dict__.setdefault('args', []).insert(0, (args, kwargs))
        return func
    return _decorator


class VmExcludeCommands(object):
    """Class for managing VM excludes"""

    EXCLUSION_TYPES = {
        'domain': 0,
        'project': 1,
        'user': 2
    }

    EXCLUSION_MAP = {
        0: 'domain',
        1: 'project',
        2: 'user'
    }

    description = "Subcommands for managing VM excludes"

    list_description = "List exclusions"

    @args('--type', metavar='<exclude-type>', dest='excludeType',
          default=None,
          help='Type of exclusion (domain, project, user)')
    def list(self, excludeType=None):
        repositories.setup_database_engine_and_factory()
        repo = repositories.get_vmexclude_repository()
        exclude_type = None
        if excludeType is not None:
            print("Filter by exclude type %s" % (excludeType))
        if excludeType == 'domain':
            exclude_type = 0
        elif excludeType == 'project':
            exclude_type = 1
        elif excludeType == 'user':
            exclude_type = 2
        excludes = repo.get_type_entities(exclude_type=exclude_type)
        headers = [
            'id',
            'type',
            'exclude_id'
        ]
        pt = prettytable.PrettyTable(headers)

        for instance in excludes:
            pt.add_row(
                [
                    instance.id,
                    VmExcludeCommands.EXCLUSION_MAP[instance.exclude_type],
                    instance.exclude_id
                ]
            )
        if six.PY3:
            print(encodeutils.safe_encode(pt.get_string()).decode())
        else:
            print(encodeutils.safe_encode(pt.get_string()))

    add_description = "Add exclusion"

    @args('--type', metavar='<exclude-type>', dest='excludeType',
          default=None,
          help='Type of exclusion (domain, project, user)')
    @args('--id', metavar='<exclude-id>', dest='excludeId',
          default=None,
          help='id of entity to exclude (domain, project, user)')
    def add(self, excludeType=None, excludeId=None):
        if excludeId is None:
            print("id option is mandatory")
            return
        if excludeType not in VmExcludeCommands.EXCLUSION_TYPES:
            print("type is not valid")
            return
        repositories.setup_database_engine_and_factory()
        repo = repositories.get_vmexclude_repository()
        entity = VmExclude()
        entity.exclude_type = VmExcludeCommands.EXCLUSION_TYPES[excludeType]
        entity.exclude_id = excludeId
        try:
            entity = repo.create_exclude(entity)
        except Exception as e:
            print(str(e))
            return
        repositories.commit()
        headers = [
            'id',
            'type',
            'exclude_id'
        ]
        pt = prettytable.PrettyTable(headers)
        pt.add_row(
            [
                entity.id,
                VmExcludeCommands.EXCLUSION_MAP[entity.exclude_type],
                entity.exclude_id
            ]
        )
        if six.PY3:
            print(encodeutils.safe_encode(pt.get_string()).decode())
        else:
            print(encodeutils.safe_encode(pt.get_string()))

    delete_description = "Delete exclusion"

    @args('--id', metavar='<exclude-id>', dest='excludeId',
          default=None,
          help='id of exclude')
    def delete(self, excludeId=None):
        if excludeId is None:
            print("Missing mandatory id parameter")
            return
        repositories.setup_database_engine_and_factory()
        repo = repositories.get_vmexclude_repository()
        repo.delete_entity_by_id(excludeId)
        repositories.commit()
        print("Exclude deleted")


class VmExpireCommands(object):
    """Class for managing VM expiration"""

    description = "Subcommands for managing VM expiration"

    list_description = "List VM expirations"

    @args('--instance', metavar='<instance-id>', dest='instanceid',
          default=None,
          help='Instance id')
    @args('--days', metavar='<expire-in>', dest='days',
          default=None,
          help='Filter VM expiring in X days')
    def list(self, instanceid=None, days=None):
        repositories.setup_database_engine_and_factory()
        repo = repositories.get_vmexpire_repository()
        res = repo.get_all_by(instance_id=instanceid, project_id=None)
        headers = [
            'id',
            'expire',
            'instance.name',
            'instance.id',
            'project.id',
            'user.id',
            'notif',
            'notif.last'
        ]
        limit = None
        if days:
            try:
                dt = datetime.datetime.now() + datetime.timedelta(days=int(days))
                limit = time.mktime(dt.timetuple())
            except Exception as e:
                print(str(e))
                return
        pt = prettytable.PrettyTable(headers)
        for instance in res:
            if limit and instance.expire > limit:
                continue

            pt.add_row(
                [
                    instance.id,
                    datetime.datetime.fromtimestamp(instance.expire),
                    instance.instance_name,
                    instance.instance_id,
                    instance.project_id,
                    instance.user_id,
                    instance.notified,
                    instance.notified_last
                ]
            )
        if six.PY3:
            print(encodeutils.safe_encode(pt.get_string()).decode())
        else:
            print(encodeutils.safe_encode(pt.get_string()))

    extend_description = "Extend a VM duration"

    @args('--id', metavar='<id>', dest='expirationid',
          help='Expiration id')
    def extend(self, expirationid):
        if not expirationid:
            print("Missing id paramerer")
            return
        repositories.setup_database_engine_and_factory()
        repo = repositories.get_vmexpire_repository()
        repo.extend_vm(entity_id=expirationid)
        repositories.commit()
        print("VM expiration successfully extended!")

    remove_description = "Deletes a VM expiration"

    @args('--id', metavar='<expiration-id>', dest='expirationid',
          help='Expiration id')
    def remove(self, expirationid):
        if not expirationid:
            print("Missing id paramerer")
            return
        repositories.setup_database_engine_and_factory()
        repo = repositories.get_vmexpire_repository()
        repo.delete_entity_by_id(entity_id=expirationid)
        repositories.commit()
        print("VM expiration successfully generated!")


CATEGORIES = {
    'vm': VmExpireCommands,
    'exclude': VmExcludeCommands
}


# Modifying similar code from nova/cmd/manage.py
def methods_of(obj):
    """Get all callable methods of an object that don't start with underscore

    returns a list of tuples of the form (method_name, method)
    """

    result = []
    for fn in dir(obj):
        if callable(getattr(obj, fn)) and not fn.startswith('_'):
            result.append((fn, getattr(obj, fn),
                          getattr(obj, fn + '_description', None)))
    return result


# Shamelessly taking same code from nova/cmd/manage.py
def add_command_parsers(subparsers):
    """Add subcommand parser to oslo_config object"""

    for category in CATEGORIES:
        command_object = CATEGORIES[category]()

        desc = getattr(command_object, 'description', None)
        parser = subparsers.add_parser(category, description=desc)
        parser.set_defaults(command_object=command_object)

        category_subparsers = parser.add_subparsers(dest='action')

        for (action, action_fn, action_desc) in methods_of(command_object):
            parser = category_subparsers.add_parser(action,
                                                    description=action_desc)

            action_kwargs = []
            for args, kwargs in getattr(action_fn, 'args', []):
                # Assuming dest is the arg name without the leading
                # hyphens if no dest is supplied
                kwargs.setdefault('dest', args[0][2:])
                if kwargs['dest'].startswith('action_kwarg_'):
                    action_kwargs.append(
                        kwargs['dest'][len('action_kwarg_'):])
                else:
                    action_kwargs.append(kwargs['dest'])
                    kwargs['dest'] = 'action_kwarg_' + kwargs['dest']

                parser.add_argument(*args, **kwargs)

            parser.set_defaults(action_fn=action_fn)
            parser.set_defaults(action_kwargs=action_kwargs)

            parser.add_argument('action_args', nargs='*',
                                help=argparse.SUPPRESS)


# Define subcommand category
category_opt = cfg.SubCommandOpt('category',
                                 title='Command categories',
                                 help='Available categories',
                                 handler=add_command_parsers)


def main():
    """Parse options and call the appropriate class/method."""
    CONF = config.new_config()
    CONF.register_cli_opt(category_opt)

    try:
        logging.register_options(CONF)
        logging.setup(CONF, "osvmexpire-manage")
        cfg_files = cfg.find_config_files(project='os-vm-expire')
        CONF(args=sys.argv[1:],
             project='os-vm-expire',
             prog='osvmexpire-manage',
             version=os_vm_expire.version.__version__,
             default_config_files=cfg_files)

    except RuntimeError as e:
        sys.exit("ERROR: %s" % e)

    # find sub-command and its arguments
    fn = CONF.category.action_fn
    fn_args = [arg.decode('utf-8') for arg in CONF.category.action_args]
    fn_kwargs = {}
    for k in CONF.category.action_kwargs:
        v = getattr(CONF.category, 'action_kwarg_' + k)
        if v is None:
            continue
        if isinstance(v, bytes):
            v = v.decode('utf-8')
        fn_kwargs[k] = v

    # call the action with the remaining arguments
    try:
        return fn(*fn_args, **fn_kwargs)
    except Exception as e:
        sys.exit("ERROR: %s" % e)


if __name__ == '__main__':
    main()



# Copyright (c) 2013-2014 Rackspace, Inc.
#               2017 IRISA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Defines interface for DB access that Resource controllers may reference
TODO: The top part of this file was 'borrowed' from Glance, but seems
quite intense for sqlalchemy, and maybe could be simplified.
"""

import datetime
import logging
import re
import threading
import time

from oslo_db import exception as db_exc
from oslo_db.sqlalchemy import session
from oslo_utils import timeutils
# from oslo_utils import uuidutils
import sqlalchemy
# from sqlalchemy import func as sa_func
# from sqlalchemy import or_
import sqlalchemy.orm as sa_orm

from os_vm_expire.common import config
from os_vm_expire.common import utils
from os_vm_expire import i18n as u
from os_vm_expire.model.migration import commands
from os_vm_expire.model import models

LOG = utils.getLogger(__name__)

_ENGINE = None
_SESSION_FACTORY = None
BASE = models.BASE
sa_logger = None

# Singleton repository references, instantiated via get_xxxx_repository()
#   functions below.  Please keep this list in alphabetical order.
_VMEXPIRE_REPOSITORY = None

CONF = config.CONF

_FACADE = None
_LOCK = threading.Lock()


def _create_facade_lazily():
    global _LOCK, _FACADE

    if _FACADE is None:
        with _LOCK:
            if _FACADE is None:
                _FACADE = session.EngineFacade.from_config(CONF,
                                                           sqlite_fk=True)
    return _FACADE


def hard_reset():
    """Performs a hard reset of database resources, used for unit testing."""
    # TODO(jvrbanac): Remove this as soon as we improve our unit testing
    # to not require this.
    global _ENGINE, _SESSION_FACTORY
    if _ENGINE:
        _ENGINE.dispose()
    _ENGINE = None
    _SESSION_FACTORY = None

    # Make sure we reinitialize the engine and session factory
    setup_database_engine_and_factory()


def setup_database_engine_and_factory():
    global sa_logger, _SESSION_FACTORY, _ENGINE

    LOG.info('Setting up database engine and session factory')
    if CONF.debug:
        sa_logger = logging.getLogger('sqlalchemy.engine')
        sa_logger.setLevel(logging.DEBUG)

    _ENGINE = _get_engine(_ENGINE)

    # Utilize SQLAlchemy's scoped_session to ensure that we only have one
    # session instance per thread.
    session_maker = sa_orm.sessionmaker(bind=_ENGINE)
    _SESSION_FACTORY = sqlalchemy.orm.scoped_session(session_maker)


def start():
    """Start for read-write requests placeholder

    Typically performed at the start of a request cycle, say for POST or PUT
    requests.
    """
    pass


def start_read_only():
    """Start for read-only requests placeholder

    Typically performed at the start of a request cycle, say for GET or HEAD
    requests.
    """
    pass


def commit():
    """Commit session state so far to the database.

    Typically performed at the end of a request cycle.
    """
    get_session().commit()


def rollback():
    """Rollback session state so far.

    Typically performed when the request cycle raises an Exception.
    """
    get_session().rollback()


def clear():
    """Dispose of this session, releases db resources.

    Typically performed at the end of a request cycle, after a
    commit() or rollback().
    """
    if _SESSION_FACTORY:  # not initialized in some unit test
        _SESSION_FACTORY.remove()


def get_session():
    """Helper method to grab session."""
    return _SESSION_FACTORY()


def _get_engine(engine):
    if not engine:
        db_connection = None
        try:
            engine = _create_engine()
            db_connection = engine.connect()
        except Exception as err:
            msg = u._("Error configuring registry database with supplied "
                      "connection. Got error: {error}").format(error=err)
            LOG.exception(msg)
            raise Exception(msg)
        finally:
            if db_connection:
                db_connection.close()

    return engine


def is_db_connection_error(args):
    """Return True if error in connecting to db."""
    # NOTE(adam_g): This is currently MySQL specific and needs to be extended
    #               to support Postgres and others.
    conn_err_codes = ('2002', '2003', '2006')
    for err_code in conn_err_codes:
        if args.find(err_code) != -1:
            return True
    return False


def _create_engine():
    LOG.debug('Sql connection')

    # engine = session.create_engine(connection, **engine_args)
    engine = _create_facade_lazily().get_engine()

    # Wrap the engine's connect method with a retry decorator.
    engine.connect = wrap_db_error(engine.connect)

    return engine


def _auto_generate_tables(engine, tables):
    if tables and 'alembic_version' in tables:
        # Upgrade the database to the latest version.
        LOG.info('Updating schema to latest version')
        commands.upgrade()
    else:
        # Create database tables from our models.
        LOG.info('Auto-creating osvmexpire registry DB')
        models.BASE.metadata.create_all(engine)

        # Sync the alembic version 'head' with current models.
        commands.stamp()


def wrap_db_error(f):
    """Retry DB connection. Copied from nova and modified."""
    def _wrap(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except sqlalchemy.exc.OperationalError as e:
            if not is_db_connection_error(e.args[0]):
                raise

            remaining_attempts = CONF.sql_max_retries
            while True:
                LOG.warning('SQL connection failed. %d attempts left.',
                            remaining_attempts)
                remaining_attempts -= 1
                time.sleep(CONF.sql_retry_interval)
                try:
                    return f(*args, **kwargs)
                except sqlalchemy.exc.OperationalError as e:
                    if (remaining_attempts <= 0 or not
                            is_db_connection_error(e.args[0])):
                        raise
                except sqlalchemy.exc.DBAPIError:
                    raise
        except sqlalchemy.exc.DBAPIError:
            raise
    _wrap.__name__ = f.__name__
    return _wrap


def delete_all_project_resources(project_id):
    """Logic to cleanup all project resources.

    This cleanup uses same alchemy session to perform all db operations as a
    transaction and will commit only when all db operations are performed
    without error.
    """
    session = get_session()

    container_repo = get_vmexpire_repository()
    container_repo.delete_project_entities(
        project_id, suppress_exception=False, session=session)


class BaseRepo(object):
    """Base repository for the osvmexpire entities.

    This class provides template methods that allow sub-classes to hook
    specific functionality as needed. Clients access instances of this class
    via singletons, therefore implementations should be stateless aside from
    configuration.
    """

    def get_session(self, session=None):
        LOG.debug("Getting session...")
        return session or get_session()

    def get_all_by(self, instance_id=None, project_id=None, session=None):
        session = self.get_session(session)
        query = session.query(models.VmExpire)
        if instance_id:
            query = query.filter_by(instance_id=instance_id)
        if project_id:
            query = query.filter_by(project_id=project_id)

        return query.all()

    def get_by_instance(self, instance_id, session=None):
        session = self.get_session(session)

        try:
            query = session.query(
                models.VmExpire
                ).filter_by(instance_id=instance_id)
            entity = query.one()

        except sa_orm.exc.NoResultFound:
            LOG.debug("Not found for %s", instance_id)
            entity = None
            _raise_entity_not_found(self._do_entity_name(), instance_id)

        return entity

    def get(self, entity_id,
            force_show_deleted=False,
            suppress_exception=False, session=None):
        """Get an entity or raise if it does not exist."""
        session = self.get_session(session)

        try:
            query = self._do_build_get_query(entity_id,
                                             session)

            entity = query.one()

        except sa_orm.exc.NoResultFound:
            LOG.debug("Not found for %s", entity_id)
            entity = None
            _raise_entity_not_found(self._do_entity_name(), entity_id)

        return entity

    def extend_vm(self, entity_id, session=None):
        session = self.get_session(session)
        try:
            query = self._do_build_get_query(entity_id,
                                             session)

            entity = query.one()

            entity.expire = int(
                time.mktime(datetime.datetime.now().timetuple()) +
                CONF.max_vm_extend * 3600 * 24
                )

            if entity.created_at + datetime.timedelta(days=CONF.max_vm_total_duration) < datetime.datetime.fromtimestamp(entity.expire):
                _raise_entity_max_extend_reached(entity.id)
                return None
            entity.notified = False
            entity.notified_last = False
            entity.save(session=session)
        except sa_orm.exc.NoResultFound:
            LOG.debug("Not found for %s", entity_id)
            entity = None
            _raise_entity_not_found(self._do_entity_name(), entity_id)
        return entity

    def create_from(self, entity, session=None):
        """Sub-class hook: create from entity."""
        if not entity:
            msg = u._(
                "Must supply non-None {entity_name}."
            ).format(entity_name=self._do_entity_name())
            raise Exception(msg)

        if entity.id:
            msg = u._(
                "Must supply {entity_name} with id=None (i.e. new entity)."
            ).format(entity_name=self._do_entity_name())
            raise Exception(msg)

        LOG.debug("Begin create from...")
        session = self.get_session(session)
        start = time.time()  # DEBUG

        # Validate the attributes before we go any further. From my
        # (unknown Glance developer) investigation, the @validates
        # decorator does not validate
        # on new records, only on existing records, which is, well,
        # idiotic.
        self._do_validate(entity.to_dict())

        try:
            LOG.debug("Saving entity...")
            entity.save(session=session)
        except db_exc.DBDuplicateEntry as e:
            session.rollback()
            LOG.exception('Problem saving entity for create')
            error_msg = re.sub('[()]', '', str(e.args))
            raise Exception(error_msg)

        LOG.debug('Elapsed repo '
                  'create vmexpire:%s', (time.time() - start))  # DEBUG

        return entity

    def save(self, entity, session=None):
        """Saves the state of the entity."""
        entity.updated_at = timeutils.utcnow()

        # Validate the attributes before we go any further. From my
        # (unknown Glance developer) investigation, the @validates
        # decorator does not validate
        # on new records, only on existing records, which is, well,
        # idiotic.
        self._do_validate(entity.to_dict())

        entity.save(session=session)

    def delete_entity_by_id(self, entity_id,
                            session=None):
        """Remove the entity by its ID."""

        session = self.get_session(session)

        entity = self.get(entity_id=entity_id,
                          session=session)

        entity.delete(session=session)

    def _do_entity_name(self):
        """Sub-class hook: return entity name, such as for debugging."""
        return "Entity"

    def _do_build_get_query(self, entity_id, external_project_id, session):
        """Sub-class hook: build a retrieve query."""
        return None

    def _do_convert_values(self, values):
        """Sub-class hook: convert text-based values to target types

        This is specifically for database values.
        """
        pass

    def _do_validate(self, values):
        """Sub-class hook: validate values.

        Validates the incoming data and raises an Invalid exception
        if anything is out of order.
        :param values: Mapping of entity metadata to check
        """
        return values

    def _update_values(self, entity_ref, values):
        for k in values:
            if getattr(entity_ref, k) != values[k]:
                setattr(entity_ref, k, values[k])

    def _build_get_project_entities_query(self, project_id, session):
        """Sub-class hook: build a query to retrieve entities for a project.

        :param project_id: id of osvmexpire project entity
        :param session: existing db session reference.
        :returns: A query object for getting all project related entities
        This will filter deleted entities if there.
        """
        msg = u._(
            "{entity_name} is missing query build method for get "
            "project entities.").format(
                entity_name=self._do_entity_name())
        raise NotImplementedError(msg)

    def get_project_entities(self, project_id, session=None):
        """Gets entities associated with a given project.

        :param project_id: id of osvmexpire project entity
        :param session: existing db session reference. If None, gets session.
        :returns: list of matching entities found otherwise returns empty list
                  if no entity exists for a given project.
        Sub-class should implement `_build_get_project_entities_query` function
        to delete related entities otherwise it would raise NotImplementedError
        on its usage.
        """

        session = self.get_session(session)
        query = self._build_get_project_entities_query(project_id, session)
        if query:
            return query.all()
        else:
            return []

    def get_entities(self, expiration_filter=None, session=None):
        """Get all entities

        :param session: existing db session reference. If None, gets session.
        :param expiration_filter: timestamp to compare expiration date with
        :returns: list of matching entities found otherwise returns empty list
                  if no entity exists for a given project.
        """

        session = self.get_session(session)
        query = session.query(models.VmExpire)
        if expiration_filter:
            query = query.filter(models.VmExpire.expire < expiration_filter)
        LOG.debug(query)
        if query:
            return query.all()
        else:
            return []

    def get_count(self, project_id, session=None):
        """Gets count of entities associated with a given project

        :param project_id: id of osvmexpire project entity
        :param session: existing db session reference. If None, gets session.
        :return: an number 0 or greater
        Sub-class should implement `_build_get_project_entities_query` function
        to delete related entities otherwise it would raise NotImplementedError
        on its usage.
        """
        session = self.get_session(session)
        query = self._build_get_project_entities_query(project_id, session)
        if query:
            return query.count()
        else:
            return 0

    def delete_project_entities(self, project_id,
                                suppress_exception=False,
                                session=None):
        """Deletes entities for a given project.

        :param project_id: id of osvmexpire project entity
        :param suppress_exception: Pass True if want to suppress exception
        :param session: existing db session reference. If None, gets session.
        Sub-class should implement `_build_get_project_entities_query` function
        to delete related entities otherwise it would raise NotImplementedError
        on its usage.
        """
        session = self.get_session(session)
        query = self._build_get_project_entities_query(project_id,
                                                       session=session)
        try:
            # query cannot be None as related repo class is expected to
            # implement it otherwise error is raised in build query call
            for entity in query:
                # Its a soft delete so its more like entity update
                entity.delete(session=session)
        except sqlalchemy.exc.SQLAlchemyError:
            LOG.exception('Problem finding project related entity to delete')
            if not suppress_exception:
                raise Exception(u._('Error deleting project '
                                    'entities for '
                                    'project_id=%s'),
                                project_id)


class VmExpireRepo(BaseRepo):
    """Repository for the expire entity."""

    def _do_entity_name(self):
        """Sub-class hook: return entity name, such as for debugging."""
        return "VMExpire"

    def _do_build_get_query(self, entity_id, session):
        """Sub-class hook: build a retrieve query."""
        query = session.query(models.VmExpire)
        query = query.filter_by(id=entity_id)
        return query

    def _do_validate(self, values):
        """Sub-class hook: validate values."""
        pass

    def _build_get_project_entities_query(self, project_id, session):
        """Builds query for retrieving orders related to given project.

        :param project_id: id of osvmexpire project entity
        :param session: existing db session reference.
        """
        return session.query(models.VmExpire).filter_by(
            project_id=project_id)

    def delete_all_entities(self, suppress_exception=False, session=None):
        """Deletes all entities.

        :param suppress_exception: Pass True if want to suppress exception
        :param session: existing db session reference. If None, gets session.
        """
        session = self.get_session(session)
        try:
            session.query(models.VmExpire).delete()
        except sqlalchemy.exc.SQLAlchemyError:
            LOG.exception('Problem deleting entities')
            if not suppress_exception:
                raise Exception(u._('Error deleting entities '))


class VmExcludeRepo(BaseRepo):
    """Repository for the exclude entity."""

    def _do_entity_name(self):
        """Sub-class hook: return entity name, such as for debugging."""
        return "VMExclude"

    def _do_build_get_query(self, entity_id, session):
        """Sub-class hook: build a retrieve query."""
        query = session.query(models.VmExclude)
        query = query.filter_by(id=entity_id)
        return query

    def _do_validate(self, values):
        """Sub-class hook: validate values."""
        pass

    def get_exclude_by_id(self, exclude_id, session=None):
        """Builds query for retrieving exclude related to given entity id.

        :param entity_id: id of entity (user, project, domain)
        :param session: existing db session reference.
        """
        session = self.get_session(session)
        return session.query(models.VmExclude).filter_by(
            exclude_id=exclude_id).one_or_none()

    def get_type_entities(self, exclude_type=None, session=None):
        """Builds query for retrieving excludes related to given type.

        :param exclude_type: id of osvmexclude type entity
        :param session: existing db session reference.
        """
        session = self.get_session(session)
        if exclude_type is None:
            return session.query(models.VmExclude).all()
        else:
            return session.query(models.VmExclude).filter_by(
                exclude_type=exclude_type).all()

    def get_exclude_type(self, exclude_name):
        """Get numeric value matching the exclude type (domain,project,user).

        :param exclude_name: domain or project or user
        :return: matching numeric value for database
        """
        if exclude_name == 'domain':
            return 0
        elif exclude_name == 'project':
            return 1
        elif exclude_name == 'user':
            return 2
        else:
            return -1

    def create_exclude_entity(self, entity_id, entity_type):
        """Get a VmExclude model

        :param entity_id: domain/project/user id
        :param entity_type: numeric value of type, see get_exclude_type()
        :return: `os_vm_expire.model.models.VmExclude`
        """
        entity = models.VmExclude()
        entity.exclude_id = entity_id
        entity.exclude_type = entity_type
        return entity

    def create_exclude(self, entity, session=None):
        """Record a new Exclude model in database.

        Checks that entity id does not already exists
        :return: `os_vm_expire.model.models.VmExclude`
        """

        ent_in_db = self.get_exclude_by_id(entity.exclude_id, session)
        if ent_in_db:
            raise Exception(u._('Error creating exclude entity '
                                'entity for '
                                'entity_id=%(id)s,'
                                'entity_type=%(type)d'
                                ' already exists') %
                            {'id': entity.exclude_id, 'type': entity.exclude_type})
        else:
            self.create_from(entity, session)
            return entity

    def delete_all_entities(self, suppress_exception=False, session=None):
        """Deletes all entities.

        :param suppress_exception: Pass True if want to suppress exception
        :param session: existing db session reference. If None, gets session.
        """
        session = self.get_session(session)
        try:
            session.query(models.VmExclude).delete()
        except sqlalchemy.exc.SQLAlchemyError:
            LOG.exception('Problem deleting entities')
            if not suppress_exception:
                raise Exception(u._('Error deleting entities '))


def get_vmexpire_repository():
    """Returns a singleton repository instance."""
    global _VMEXPIRE_REPOSITORY
    return _get_repository(_VMEXPIRE_REPOSITORY, VmExpireRepo)


def get_vmexclude_repository():
    """Returns a singleton repository instance."""
    global _VMEXPIRE_REPOSITORY
    return _get_repository(_VMEXPIRE_REPOSITORY, VmExcludeRepo)


def _get_repository(global_ref, repo_class):
    if not global_ref:
        global_ref = repo_class()
    return global_ref


def _raise_entity_max_extend_reached(entity_id):
    raise Exception(u._("VM reached its maximum life, {id}, cannot extend it").format(
        id=entity_id))


def _raise_entity_not_found(entity_name, entity_id):
    raise Exception(u._("No {entity} found with ID {id}").format(
        entity=entity_name,
        id=entity_id))


def _raise_entity_id_not_found(entity_id):
    raise Exception(u._("Entity ID {entity_id} not "
                        "found").format(entity_id=entity_id))


def _raise_no_entities_found(entity_name):
    raise Exception(
        u._("No entities of type {entity_name} found").format(
            entity_name=entity_name))



#!/usr/bin/env python

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Osvmexpire worker server.
"""
import datetime
from email.mime.text import MIMEText
import eventlet
import os
import smtplib
import sys
import time


from os_vm_expire.common import config
from os_vm_expire.common import utils
from os_vm_expire.model import repositories
from os_vm_expire import version

from oslo_log import log
from oslo_service import service

# import futurist
from futurist import periodics
import requests

# Oslo messaging RPC server uses eventlet.
eventlet.monkey_patch()

# 'Borrowed' from the Glance project:
# If ../os_vm_expire/__init__.py exists, add ../ to Python search path, so that
# it will override what happens to be installed in /usr/(local/)lib/python...
possible_topdir = os.path.normpath(os.path.join(os.path.abspath(sys.argv[0]),
                                   os.pardir,
                                   os.pardir))
base_local_dir = os.path.join(possible_topdir, 'os_vm_expire', '__init__.py')
if os.path.exists(base_local_dir):
    sys.path.insert(0, possible_topdir)


LOG = utils.getLogger(__name__)


def fail(returncode, e):
    sys.stderr.write("ERROR: {0}\n".format(e))
    sys.exit(returncode)


def delete_vm(instance_id, project_id, token):
    '''Delete a VM on expiration

    conf in [cleaner]
    nova_url = http://controller.genouest.org:8774/v2.1/%(tenant_id)s
    '''
    nova_url = config.CONF.cleaner.nova_url % {
        'tenant_id': project_id,
        'project_id': project_id
        }
    LOG.debug('Nova URI:' + nova_url)
    headers = {
        'X-Auth-Token': token,
        'Content-Type': 'application/json',
        'Accept': 'application/json'
    }
    r = requests.delete(nova_url + '/servers/' + instance_id, headers=headers)
    if r.status_code != 204:
        LOG.error('Failed to delete instance ' + str(instance_id))
        return False
    else:
        LOG.info('DELETE:' + str(instance_id) + ':' + str(project_id))
        return True


def get_identity_token():
    conf_cleaner = config.CONF.cleaner
    ks_uri = conf_cleaner.auth_uri

    auth = {
        'auth': {
            'scope':
                {'project': {
                    'name': conf_cleaner.admin_service,
                    'domain':
                        {
                            'name': conf_cleaner.admin_project_domain_name
                        }
                    }
                 },
            'identity': {
                    'password': {
                        'user': {
                            'domain': {
                                'name': conf_cleaner.admin_user_domain_name
                            },
                            'password': conf_cleaner.admin_password,
                            'name': conf_cleaner.admin_user
                        }
                    },
                    'methods': ['password']
                }
        }
    }
    r = requests.post(ks_uri + '/auth/tokens', json=auth)
    if 'X-Subject-Token' not in r.headers:
        LOG.error('Could not get authorization')
        return None
    token = r.headers['X-Subject-Token']
    return token


def get_project_name(project_id, token):
    LOG.debug("Get project name")
    # fetch user from identity to get user email
    headers = {
        'X-Auth-Token': token,
        'Content-Type': 'application/json',
        'Accept': 'application/json'
    }
    ks_uri = config.CONF.cleaner.auth_uri
    try:
        r = requests.get(ks_uri + '/projects/' + project_id, headers=headers)
    except Exception:
        LOG.exception('Failed to get project name for id ' + str(project_id))
        return None
    if r.status_code != 200:
        return None
    project = r.json()
    if 'project' in project:
        return project['project']['name']
    return None


def send_email(instance, token, delete=False):
    LOG.debug("Send expiration notification mail")
    # fetch user from identity to get user email
    headers = {
        'X-Auth-Token': token,
        'Content-Type': 'application/json',
        'Accept': 'application/json'
    }
    ks_uri = config.CONF.cleaner.auth_uri
    r = requests.get(ks_uri + '/users/' + instance.user_id, headers=headers)
    if r.status_code != 200:
        return False
    user = r.json()
    email = None
    if 'user' in user and 'email' in user['user']:
        email = user['user']['email']
    if email is None:
        LOG.error('Could not get email for user ' + instance.user_id)
        return False
    # send email
    to = [email]
    if (not delete) and config.CONF.smtp.email_smtp_copy_expire_notif_to is not None:
        to = [email, config.CONF.smtp.email_smtp_copy_delete_notif_to]
    if delete and config.CONF.smtp.email_smtp_copy_delete_notif_to is not None:
        to = [email, config.CONF.smtp.email_smtp_copy_delete_notif_to]

    project_name = get_project_name(instance.project_id, token)

    if project_name is None:
        project_name = instance.project_id

    subject = 'VM %s [project: %s] expiration' % (
        str(instance.instance_name),
        project_name
    )

    message = ('VM %s (id: %s, project: %s) will expire at %s,' +
               'connect to dashboard to extend its duration else ' +
               ' it will be deleted.') % (
        instance.instance_name,
        instance.instance_id,
        instance.project_id,
        str(datetime.datetime.fromtimestamp(instance.expire))
        )
    if delete:
        message = ('VM %s (id: %s, project: %s) has expired at %s' +
                   ' and has been deleted.') % (
            instance.instance_name,
            instance.instance_id,
            instance.project_id,
            str(datetime.datetime.fromtimestamp(instance.expire))
            )
    LOG.info('NOTIF:' + instance.id + ':' + str(message))
    # Create a text/plain message
    msg = MIMEText(message)

    msg['Subject'] = subject
    msg['From'] = config.CONF.smtp.email_smtp_from
    if msg['From'] is None:
        LOG.error('Missing smtp.email_smtp_from in config')
        return False
    msg['To'] = ', '.join(to)

    # Send the message via our own SMTP server, but don't include the
    # envelope header.
    try:
        s = smtplib.SMTP(config.CONF.smtp.email_smtp_host,
                         config.CONF.smtp.email_smtp_port)
        if config.CONF.smtp.email_smtp_tls:
            s.starttls()
        if config.CONF.smtp.email_smtp_user:
            s.login(config.CONF.smtp.email_smtp_user,
                    config.CONF.smtp.email_smtp_password)
        s.sendmail(msg['From'], to, msg.as_string())
        s.quit()
    except Exception:
        LOG.error('Failed to send expiration notification mail to ' + email)
        return False

    return True


# Every hour
@periodics.periodic(3600)
def check(started_at):
    token = get_identity_token()
    conf_cleaner = config.CONF.cleaner
    LOG.debug("check instances")
    repo = repositories.get_vmexpire_repository()
    now = int(time.mktime(datetime.datetime.now().timetuple()))
    check_time = now + (conf_cleaner.notify_before_days * 3600 * 24)
    last_check_time = now + (conf_cleaner.notify_before_days_last * 3600 * 24)
    entities = repo.get_entities(expiration_filter=check_time)
    for entity in entities:
        if entity.expire < check_time and not entity.notified:
            # notify
            LOG.debug("First expiration notification %s" % (entity.id))
            res = send_email(entity, token, delete=False)
            if res:
                entity.notified = True
                try:
                    entity.save()
                    repositories.commit()
                except Exception as e:
                    LOG.exception("expiration save error: " + str(e))
                    repositories.rollback()
        elif entity.expire < last_check_time and not entity.notified_last:
            # notify_last
            LOG.debug("Last expiration notification %s" % (entity.id))
            res = send_email(entity, token, delete=False)
            if res:
                entity.notified_last = True
                try:
                    entity.save()
                    repositories.commit()
                except Exception as e:
                    LOG.exception("expiration save error: " + str(e))
                    repositories.rollback()
        elif entity.expire < now:
            # delete
            LOG.debug("Delete VM %s" % (entity.id))
            res = delete_vm(entity.instance_id, entity.project_id, token)
            if res:
                try:
                    repo.delete_entity_by_id(entity_id=entity.id)
                    repositories.commit()
                except Exception as e:
                    LOG.exception("expiration deletion error: " + str(e))
                    repositories.rollback()
            send_email(entity, token, delete=True)


class CleanerServer(service.Service):

    def __init__(self):
        super(CleanerServer, self).__init__()
        repositories.setup_database_engine_and_factory()
        started_at = time.time()
        callables = [(check, (started_at,), {})]
        self.w = periodics.PeriodicWorker(callables)

    def start(self):
        LOG.info("Starting the CleanerServer")
        self.w.start()
        super(CleanerServer, self).start()

    def stop(self):
        LOG.info("Halting the CleanerServer")
        self.w.stop()
        super(CleanerServer, self).stop()


def main():
    try:
        config.CONF(sys.argv[1:],
                    project='os-vm-expire',
                    version=version.version_info.version_string)

        # Import and configure logging.
        log.setup(config.CONF, 'osvmexpire')
        LOG = log.getLogger(__name__)
        LOG.debug("Booting up os-vm-expire cleaner node...")

        service.launch(
            config.CONF,
            CleanerServer(),
            workers=1
        ).wait()

    except RuntimeError as e:
        fail(1, e)


if __name__ == '__main__':
    main()



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from os_vm_expire.model import models

from os_vm_expire.model import repositories
from os_vm_expire.tests import utils


class WhenTestingVmExcludesResource(utils.OsVMExpireAPIBaseTestCase):

    def setUp(self):
        super(WhenTestingVmExcludesResource, self).setUp()

    def tearDown(self):
        super(WhenTestingVmExcludesResource, self).tearDown()
        repo = repositories.get_vmexpire_repository()
        repo.delete_all_entities()
        repositories.commit()
        repo = repositories.get_vmexclude_repository()
        repo.delete_all_entities()
        repositories.commit()

    def test_can_get_vmexcludes(self):
        entity = create_vmexclude_model(exclude_type=0)
        create_vmexclude(entity)
        exclude_id = entity.exclude_id
        exclude_type = entity.exclude_type
        _get_resp = self.app.get('/12345project/vmexcludes/')
        self.assertEqual(200, _get_resp.status_int)
        self.assertIn('vmexcludes', _get_resp.json)
        self.assertEqual(len(_get_resp.json['vmexcludes']), 1)
        self.assertEqual(
            _get_resp.json['vmexcludes'][0]['exclude_id'],
            exclude_id
            )
        self.assertEqual(
            _get_resp.json['vmexcludes'][0]['exclude_type'],
            exclude_type
            )

    def test_can_get_vmexclude(self):
        entity = create_vmexclude_model()
        instance = create_vmexclude(entity)
        _get_resp = self.app.get(
            '/12345project/vmexcludes/' + instance.id
            )
        self.assertEqual(200, _get_resp.status_int)
        self.assertIn('vmexclude', _get_resp.json)
        self.assertEqual(
            _get_resp.json['vmexclude']['exclude_id'],
            entity.exclude_id)

    def test_can_create_vmexclude(self):
        entity = create_vmexclude_model(exclude_type=1)
        _get_resp = self.app.post_json(
            '/12345/vmexcludes/',
            {
                'id': entity.exclude_id,
                'type': 'project'
            },
            headers={'Content-Type': 'application/json'}
            )
        self.assertEqual(202, _get_resp.status_int)
        self.assertIn('vmexclude', _get_resp.json)
        self.assertEqual(
            _get_resp.json['vmexclude']['exclude_id'],
            entity.exclude_id
            )
        self.assertEqual(
            _get_resp.json['vmexclude']['exclude_type'],
            entity.exclude_type
            )
        _get_existing_resp = self.app.get(
            '/12345project/vmexcludes/' + _get_resp.json['vmexclude']['id']
        )
        self.assertEqual(200, _get_existing_resp.status_int)
        self.assertIn('vmexclude', _get_existing_resp.json)
        self.assertEqual(
            _get_resp.json['vmexclude']['exclude_id'],
            entity.exclude_id
            )

    def test_can_delete_vmexclude(self):
        entity = create_vmexclude_model()
        instance = create_vmexclude(entity)
        _get_resp = self.app.delete(
            '/12345project/vmexcludes/' + instance.id,
            headers={'Content-Type': 'application/json'}
            )
        self.assertEqual(204, _get_resp.status_int)
        _get_resp = self.app.get('/12345project/vmexcludes/')
        self.assertEqual(200, _get_resp.status_int)
        self.assertIn('vmexcludes', _get_resp.json)
        self.assertEqual(len(_get_resp.json['vmexcludes']), 0)


def create_vmexclude_model(prefix=None, exclude_type=0):
    if not prefix:
        prefix = '12345'
    entity = models.VmExclude()
    entity.exclude_id = prefix
    entity.exclude_type = exclude_type

    return entity


def create_vmexclude(entity):
    repo = repositories.get_vmexclude_repository()
    instance = repo.create_exclude(entity)
    repositories.commit()
    return instance



# Copyright (c) 2013-2014 Rackspace, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This test module focuses on RBAC interactions with the API resource classes.
For typical-flow business logic tests of these classes, see the
'resources_test.py' module.
"""
import os

import mock
from oslo_policy import policy
from webob import exc

from os_vm_expire.api.controllers import versions
from os_vm_expire.api.controllers import vmexpire
from os_vm_expire.common import config
from os_vm_expire import context
# from os_vm_expire.model import models
from os_vm_expire.tests import utils

import webob

# Point to the policy.json file located in source control.
TEST_VAR_DIR = os.path.abspath(os.path.join(os.path.dirname(__file__),
                                            '../../../etc', 'oslo-config-generator'))

TEST_CONFIG_FILE = os.path.join(TEST_VAR_DIR, 'osvmexpire.conf.sample')
TEST_POLICY_FILE = os.path.join(TEST_VAR_DIR, 'policy.json.sample')
CONF = config.new_config()

ENFORCER = policy.Enforcer(CONF, policy_file=TEST_POLICY_FILE)


class TestableResource(object):

    def __init__(self, *args, **kwargs):
        self.controller = self.controller_cls(*args, **kwargs)

    def on_get(self, req, resp, *args, **kwargs):
        with mock.patch('pecan.request', req):
            with mock.patch('pecan.response', resp):
                return self.controller.on_get(*args, **kwargs)

    def on_post(self, req, resp, *args, **kwargs):
        with mock.patch('pecan.request', req):
            with mock.patch('pecan.response', resp):
                return self.controller.on_post(*args, **kwargs)

    def on_put(self, req, resp, *args, **kwargs):
        with mock.patch('pecan.request', req):
            with mock.patch('pecan.response', resp):
                return self.controller.on_put(*args, **kwargs)

    def on_delete(self, req, resp, *args, **kwargs):
        with mock.patch('pecan.request', req):
            with mock.patch('pecan.response', resp):
                return self.controller.on_delete(*args, **kwargs)


class VersionsResource(TestableResource):
    controller_cls = versions.VersionsController


class VmExpireResource(TestableResource):
    controller_cls = vmexpire.VmExpireController


class BaseTestCase(utils.OsVMExpireAPIBaseTestCase, utils.MockModelRepositoryMixin):

    def setUp(self):
        super(BaseTestCase, self).setUp()
        CONF(args=['--config-dir', TEST_VAR_DIR])
        CONF(args=['--config-file', TEST_CONFIG_FILE])
        self.policy_enforcer = ENFORCER
        self.policy_enforcer.load_rules(True)
        self.resp = mock.MagicMock()

    def _generate_req(self, roles=None, accept=None, content_type=None,
                      user_id=None, project_id=None):
        """Generate a fake HTTP request with security context added to it."""
        req = mock.MagicMock()
        req.get_param.return_value = None

        kwargs = {
            'user': user_id,
            'project': project_id,
            'roles': roles or [],
            'policy_enforcer': self.policy_enforcer,
        }
        req.environ = {}
        req.environ['os_vm_expire.context'] = context.RequestContext(**kwargs)
        req.content_type = content_type
        if accept:
            req.accept.header_value.return_value = accept
        else:
            req.accept = None

        return req

    def _generate_stream_for_exit(self):
        """Mock HTTP stream generator, to force RBAC-pass exit.

        Generate a fake HTTP request stream that forces an IOError to
        occur, which short circuits API resource processing when RBAC
        checks under test here pass.
        """
        stream = mock.MagicMock()
        read = mock.MagicMock(return_value=None, side_effect=IOError())
        stream.read = read
        return stream

    def _assert_post_rbac_exception(self, exception, role):
        """Assert that we received the expected RBAC-passed exception."""
        self.assertEqual(500, exception.status_int)

    def _generate_get_error(self):
        """Falcon exception generator to throw from early-exit mocks.

        Creates an exception that should be raised by GET tests that pass
        RBAC. This allows such flows to short-circuit normal post-RBAC
        processing that is not tested in this module.
        :return: Python exception that should be raised by repo get methods.
        """
        # The 'Read Error' clause needs to match that asserted in
        #    _assert_post_rbac_exception() above.
        return exc.HTTPServerError(detail='Read Error')

    def _assert_pass_rbac(self, roles, method_under_test, accept=None,
                          content_type=None, user_id=None, project_id=None):
        """Assert that RBAC authorization rules passed for the specified roles.

        :param roles: List of roles to check, one at a time
        :param method_under_test: The test method to invoke for each role.
        :param accept Optional Accept header to set on the HTTP request
        :return: None
        """
        for role in roles:
            self.req = self._generate_req(roles=[role] if role else [],
                                          accept=accept,
                                          content_type=content_type,
                                          user_id=user_id,
                                          project_id=project_id)
            # Force an exception early past the RBAC passing.
            type(self.req).body = mock.PropertyMock(side_effect=IOError)
            self.req.body_file = self._generate_stream_for_exit()

            exception = self.assertRaises(exc.HTTPServerError,
                                          method_under_test)
            self._assert_post_rbac_exception(exception, role)

    def _assert_fail_rbac(self, roles, method_under_test, accept=None,
                          content_type=None, user_id=None, project_id=None):
        """Assert that RBAC rules failed for one of the specified roles.

        :param roles: List of roles to check, one at a time
        :param method_under_test: The test method to invoke for each role.
        :param accept Optional Accept header to set on the HTTP request
        :return: None
        """
        for role in roles:
            self.req = self._generate_req(roles=[role] if role else [],
                                          accept=accept,
                                          content_type=content_type,
                                          user_id=user_id,
                                          project_id=project_id)

            exception = self.assertRaises(exc.HTTPForbidden, method_under_test)
            self.assertEqual(403, exception.status_int)


class WhenTestingVersionsResource(BaseTestCase):
    """RBAC tests for the barbican.api.resources.VersionsResource class."""
    def setUp(self):
        super(WhenTestingVersionsResource, self).setUp()

        self.resource = VersionsResource()

    def test_rules_should_be_loaded(self):
        self.assertIsNotNone(self.policy_enforcer.rules)

    def test_should_pass_get_versions(self):
        # Can't use base method that short circuits post-RBAC processing here,
        # as version GET is trivial
        for role in ['admin', 'member']:
            self.req = self._generate_req(roles=[role] if role else [])
            self._invoke_on_get()

    def test_should_pass_get_versions_with_bad_roles(self):
        self.req = self._generate_req(roles=[None, 'bunkrolehere'])
        self._invoke_on_get()

    def test_should_pass_get_versions_with_no_roles(self):
        self.req = self._generate_req()
        self._invoke_on_get()

    def test_should_pass_get_versions_multiple_roles(self):
        self.req = self._generate_req(roles=['admin', 'member'])
        self._invoke_on_get()

    def _invoke_on_get(self):
        self.resource.on_get(self.req, self.resp)


class WhenTestingVmExpireResource(BaseTestCase):
    """RBAC tests for the barbican.api.resources.SecretsResource class."""
    def setUp(self):
        super(WhenTestingVmExpireResource, self).setUp()

        self.vmexpire_repo = mock.MagicMock()
        self.setup_vmexpire_repository_mock(self.vmexpire_repo)

        self.project_id = '12345project'
        self.user_id = '12345user'
        self.resource = VmExpireResource(self.project_id)

    def test_rules_should_be_loaded(self):
        self.assertIsNotNone(self.policy_enforcer.rules)

    def test_should_pass_get_vmexpires(self):
        self.req = self._generate_req(roles=['member'],
            content_type='application/json',
            project_id=self.project_id)
        res = self._invoke_on_get()
        self.assertIsNotNone(res)

    def test_should_pass_get_vmexpire(self):
        self.req = self._generate_req(roles=['member'],
            content_type='application/json',
            project_id=self.project_id)
        res = self._invoke_on_get('12345expire')
        self.assertIsNotNone(res)

    def test_should_fail_get_vmexpires(self):
        self.req = self._generate_req(roles=['bogus'])
        self.assertRaises(webob.exc.HTTPForbidden, self._invoke_on_get)

    def test_should_fail_get_vmexpire(self):
        self.req = self._generate_req(roles=['bogus'])
        self.assertRaises(webob.exc.HTTPForbidden, self._invoke_on_get, '12345')

    def test_should_pass_get_vmexpires_again(self):
        self.req = self._generate_req(roles=['admin', 'member'])
        res = self._invoke_on_get()
        self.assertIsNotNone(res)

    def test_should_fail_put_vmexpire(self):
        self.req = self._generate_req(roles=['bogus'])
        self.assertRaises(webob.exc.HTTPForbidden, self._invoke_on_put, '12345')

    def test_should_admin_pass_put_vmexpire(self):
        self.req = self._generate_req(roles=['admin'],
            content_type='application/json')
        res = self._invoke_on_put('12345expire')
        self.assertIsNotNone(res)

    def test_should_user_pass_put_vmexpire(self):
        self.req = self._generate_req(roles=['member'],
            content_type='application/json',
            project_id=self.project_id)
        res = self._invoke_on_put('12345expire')
        self.assertIsNotNone(res)

    def test_should_fail_delete_vmexpire(self):
        self.req = self._generate_req(roles=['member'],
            content_type='application/json',
            project_id=self.project_id)
        self.assertRaises(webob.exc.HTTPForbidden, self._invoke_on_delete, '12345')

    def _invoke_on_put(self, instance_id=None):
        return self.resource.on_put(self.req, self.resp, 'vmexpires', instance_id)

    def _invoke_on_delete(self, instance_id=None):
        return self.resource.on_delete(self.req, self.resp, 'vmexpires', instance_id)

    def _invoke_on_get(self, instance_id=None):
        return self.resource.on_get(self.req, self.resp, 'vmexpires', instance_id)



# Copyright (c) 2016 IBM
#               2017 IRISA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from os_vm_expire.common import config
from os_vm_expire.model import models
from os_vm_expire.model import repositories as repo
from oslo_log import log
from oslo_utils import timeutils

# from sqlalchemy import sql as sa_sql

import datetime

# Import and configure logging.
CONF = config.CONF
log.setup(CONF, 'osvmexpire')
LOG = log.getLogger(__name__)


def cleanup_softdeletes(model, threshold_date=None):
    """Remove soft deletions from a table.

    :param model: table class to remove soft deletions
    :param threshold_date: soft deletions older than this date will be removed
    :returns: total number of entries removed from the database
    """
    LOG.debug("Cleaning soft deletes: %s", model.__name__)
    session = repo.get_session()
    query = session.query(model)
    query = query.filter_by(deleted=True)
    if threshold_date:
        query = query.filter(model.deleted_at <= threshold_date)
    delete_count = query.delete()
    LOG.info("Cleaned up %(delete_count)s entries for %(model_name)s",
             {'delete_count': delete_count,
              'model_name': model.__name__})
    return delete_count


def cleanup_all(threshold_date=None):
    """Clean up the main soft deletable resources.

    This function contains an order of calls to
    clean up the soft-deletable resources.
    :param threshold_date: soft deletions older than this date will be removed
    :returns: total number of entries removed from the database
    """
    LOG.debug("Cleaning up soft deletions where deletion date"
              " is older than %s", str(threshold_date))
    total = 0
    total += cleanup_softdeletes(models.VmExpire,
                                 threshold_date=threshold_date)

    LOG.info("Cleaned up %s soft deleted entries", total)
    return total


def clean_command(sql_url, min_num_days, verbose, log_file):
    """Clean command to clean up the database.

    :param sql_url: sql connection string to connect to a database
    :param min_num_days: clean up soft deletions older than this date
    :param verbose: If True, log and print more information
    :param log_file: If set, override the log_file configured
    """
    if verbose:
        # The verbose flag prints out log events to the screen, otherwise
        # the log events will only go to the log file
        CONF.set_override('debug', True)

    if log_file:
        CONF.set_override('log_file', log_file)

    LOG.info("Cleaning up soft deletions in the barbican database")
    log.setup(CONF, 'osvmexpire')

    cleanup_total = 0
    current_time = timeutils.utcnow()
    stop_watch = timeutils.StopWatch()
    stop_watch.start()
    try:
        if sql_url:
            CONF.set_override('sql_connection', sql_url)
        repo.setup_database_engine_and_factory()

        threshold_date = None
        if min_num_days >= 0:
            threshold_date = current_time - datetime.timedelta(
                days=min_num_days)
        else:
            threshold_date = current_time
        cleanup_total += cleanup_all(threshold_date=threshold_date)
        repo.commit()

    except Exception as ex:
        LOG.exception('Failed to clean up soft deletions in database.')
        repo.rollback()
        cleanup_total = 0  # rollback happened, no entries affected
        raise ex
    finally:
        stop_watch.stop()
        elapsed_time = stop_watch.elapsed()
        if verbose:
            CONF.clear_override('debug')

        if log_file:
            CONF.clear_override('log_file')
        repo.clear()

        if sql_url:
            CONF.clear_override('sql_connection')

        log.setup(CONF, 'osvmexpire')  # reset the overrides

        LOG.info("Cleaning of database affected %s entries", cleanup_total)
        LOG.info('DB clean up finished in %s seconds', elapsed_time)



# Copyright (c) 2013-2014 Rackspace, Inc.
#               2017 IRISA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Defines database models for osvmexpire
"""
from oslo_serialization import jsonutils as json
from oslo_utils import timeutils
import six
import sqlalchemy as sa
from sqlalchemy.ext import compiler
from sqlalchemy.ext import declarative
from sqlalchemy import orm
# from sqlalchemy.orm import collections as col
from sqlalchemy import types as sql_types

from os_vm_expire.common import utils
# from os_vm_expire import i18n as u

BASE = declarative.declarative_base()
ERROR_REASON_LENGTH = 255
SUB_STATUS_LENGTH = 36
SUB_STATUS_MESSAGE_LENGTH = 255


@compiler.compiles(sa.BigInteger, 'sqlite')
def compile_big_int_sqlite(type_, compiler, **kw):
    return 'INTEGER'


class JsonBlob(sql_types.TypeDecorator):
    """JsonBlob is custom type for fields which need to store JSON text."""
    impl = sa.Text

    def process_bind_param(self, value, dialect):
        if value is not None:
            return json.dumps(value)
        return value

    def process_result_value(self, value, dialect):
        if value is not None:
            return json.loads(value)
        return value


class ModelBase(object):
    """Base class for Nova and osvmexpire Models."""
    __table_args__ = {'mysql_engine': 'InnoDB'}
    __table_initialized__ = False
    __protected_attributes__ = {
        "created_at", "updated_at", "deleted_at", "deleted"}

    id = sa.Column(sa.String(36), primary_key=True,
                   default=utils.generate_uuid)

    created_at = sa.Column(sa.DateTime, default=timeutils.utcnow,
                           nullable=False)
    updated_at = sa.Column(sa.DateTime, default=timeutils.utcnow,
                           nullable=False, onupdate=timeutils.utcnow)
    deleted_at = sa.Column(sa.DateTime)
    deleted = sa.Column(sa.Boolean, nullable=False, default=False)

    def save(self, session=None):
        """Save this object."""
        # import api here to prevent circular dependency problem
        import os_vm_expire.model.repositories
        session = session or os_vm_expire.model.repositories.get_session()
        # if model is being created ensure that created/updated are the same
        if self.id is None:
            self.created_at = timeutils.utcnow()
            self.updated_at = self.created_at
        session.add(self)
        session.flush()

    def delete(self, session=None):
        """Delete this object."""
        import os_vm_expire.model.repositories
        session = session or os_vm_expire.model.repositories.get_session()
        self._do_delete_children(session)
        session.delete(self)

    def _do_delete_children(self, session):
        """Sub-class hook: delete children relationships."""
        pass

    def update(self, values):
        """dict.update() behaviour."""
        for k, v in values.items():
            self[k] = v

    def __setitem__(self, key, value):
        setattr(self, key, value)

    def __getitem__(self, key):
        return getattr(self, key)

    def __iter__(self):
        self._i = iter(orm.object_mapper(self).sa.Columns)
        return self

    def next(self):
        n = next(self._i).name
        return n, getattr(self, n)

    def keys(self):
        return self.__dict__.keys()

    def values(self):
        return self.__dict__.values()

    def items(self):
        return self.__dict__.items()

    def to_dict(self):
        return self.__dict__.copy()

    def to_dict_fields(self):
        """Returns a dictionary of just the db fields of this entity."""

        if self.created_at:
            created_at = self.created_at.isoformat()
        else:
            created_at = self.created_at

        if self.updated_at:
            updated_at = self.updated_at.isoformat()
        else:
            updated_at = self.updated_at

        dict_fields = {
            'created': created_at,
            'updated': updated_at,
        }

        if self.deleted_at:
            dict_fields['deleted_at'] = self.deleted_at.isoformat()
        if self.deleted:
            dict_fields['deleted'] = True
        dict_fields.update(self._do_extra_dict_fields())
        return dict_fields

    def _do_extra_dict_fields(self):
        """Sub-class hook method: return dict of fields."""
        return {}

    def _iso_to_datetime(self, expiration):
        """Convert ISO formatted string to datetime."""
        if isinstance(expiration, six.string_types):
            expiration_iso = timeutils.parse_isotime(expiration.strip())
            expiration = timeutils.normalize_time(expiration_iso)

        return expiration


class SoftDeleteMixIn(object):
    """Mix-in class that adds soft delete functionality."""

    def delete(self, session=None):
        """Delete this object."""
        import os_vm_expire.model.repositories
        session = session or os_vm_expire.model.repositories.get_session()
        self.deleted = True
        self.deleted_at = timeutils.utcnow()
        self.save(session=session)

        self._do_delete_children(session)


class VmExpire(BASE, ModelBase):
    """Represents a VM."""

    __tablename__ = 'vmexpire'

    instance_id = sa.Column(
        sa.String(255), index=True,
        nullable=False)
    project_id = sa.Column(
        sa.String(255), index=True,
        nullable=False)
    user_id = sa.Column(
        sa.String(255), index=False,
        nullable=False)
    expire = sa.Column(
        sa.Integer, index=False,
        nullable=False)
    notified = sa.Column(
        sa.Boolean, index=False,
        nullable=False)
    notified_last = sa.Column(
        sa.Boolean, index=False,
        nullable=False)
    instance_name = sa.Column(
        sa.String(255), index=False,
        nullable=True)

    __table_args__ = (sa.UniqueConstraint('instance_id',
                                          name='_vmexpire_uc'),)

    def __init__(self, parsed_request=None):
        """Creates secret from a dict."""
        super(VmExpire, self).__init__()

    def _do_extra_dict_fields(self):
        """Sub-class hook method: return dict of fields."""
        return {
            'id': self.id,
            'instance_id': self.instance_id,
            'project_id': self.project_id,
            'expire': self.expire,
            'notified': self.notified,
            'notified_last': self.notified_last,
            'user_id': self.user_id,
            'instance_name': self.instance_name
        }


class VmExclude(BASE, ModelBase):
    """Represents a VM."""

    __tablename__ = 'vmexclude'

    exclude_id = sa.Column(
        sa.String(255), index=False,
        nullable=False)
    exclude_type = sa.Column(
        sa.Integer, index=False,
        nullable=False)

    def __init__(self, parsed_request=None):
        """Creates secret from a dict."""
        super(VmExclude, self).__init__()

    def _do_extra_dict_fields(self):
        """Sub-class hook method: return dict of fields."""
        return {
            'id': self.id,
            'exclude_id': self.exclude_id,
            'exclude_type': self.exclude_type
        }



#!/usr/bin/env python

# Copyright  2017 IRISA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import os
from paste import deploy
from paste import httpserver


def run():
    vmexpire_port = '9411'
    if 'OSVMEXPIRE_PORT' in os.environ:
        vmexpire_port = os.environ['OSVMEXPIRE_PORT']

    prop_dir = 'etc/os-vm-expire'
    if not os.path.exists('/etc/os-vm-expire/osvmexpire-api-paste.ini'):
        application = deploy.loadapp(
            'config:{prop_dir}/osvmexpire-api-paste.ini'.
            format(prop_dir=prop_dir), name='main', relative_to='.')
    else:
        application = deploy.loadapp(
            'config:{prop_dir}/osvmexpire-api-paste.ini'.
            format(prop_dir=prop_dir), name='main', relative_to='/')

    httpserver.serve(application, host='0.0.0.0', port=vmexpire_port)


if __name__ == '__main__':
    run()



# Copyright (c) 2013-2014 Rackspace, Inc.
#               2017 O. Sallou, IRISA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Configuration setup for os-vm-expire.
"""

import logging
import os

from oslo_config import cfg
from oslo_log import log
from oslo_middleware import cors
from oslo_service import _options

from os_vm_expire import i18n as u
import os_vm_expire.version


MAX_VM_DURATION_DAYS = 60
MAX_VM_EXTEND_DAYS = 30
MAX_VM_TOTAL_DURATION_DAYS = 365

KS_NOTIFICATIONS_GRP_NAME = "nova_notifications"

mail_opt_group = cfg.OptGroup(name='smtp',
                              title='SMTP mail Options')

mail_opts = [
    cfg.StrOpt('email_smtp_host',
               default='localhost',
               help=u._("SMTP hostname")),
    cfg.IntOpt('email_smtp_port',
               default=25,
               help=u._("SMTP port")),
    cfg.BoolOpt('email_smtp_tls',
                default=False,
                help=u._("SMTP tls use?")),
    cfg.StrOpt('email_smtp_user',
               default=None,
               help=u._("SMTP user")),
    cfg.StrOpt('email_smtp_password',
               default=None,
               help=u._("SMTP password")),
    cfg.StrOpt('email_smtp_from',
               default=None,
               help=u._("SMTP From mail origin")),
    cfg.StrOpt('email_smtp_copy_delete_notif_to',
               default=None,
               help=u._('Email where deletion notifications should be sent,'
                        ' leave empty if no copy is needed')),
]

cleaner_opt_group = cfg.OptGroup(name='cleaner',
                                 title='Cleaner Application Options')

cleaner_opts = [
    cfg.StrOpt('auth_uri',
               default='http://controller:5000/v3.0',
               help=u._("Openstack identity url")),
    cfg.StrOpt('nova_url',
               default='http://controller:8774/v2.1',
               help=u._("Openstack nova compute url")),
    cfg.StrOpt('admin_service',
               default='service',
               help=u._("service project name")),
    cfg.StrOpt('admin_user',
               default='os_vm_expire',
               help=u._("os-vm-expire service user id")),
    cfg.StrOpt('admin_password',
               help=u._("os-vm-expire service user password")),
    cfg.StrOpt('admin_user_domain_name',
               default='default',
               help=u._("os-vm-expire user domain name")),
    cfg.StrOpt('admin_project_domain_name',
               default='default',
               help=u._("os-vm-expire service project domain name")),
    cfg.IntOpt('notify_before_days',
               default=10,
               help=u._("os-vm-expire send notification before X days")),
    cfg.IntOpt('notify_before_days_last',
               default=2,
               help=u._("os-vm-expire send last notification before X days")),
]


worker_opt_group = cfg.OptGroup(name='worker',
                                title='Worker Application Options')

worker_opts = [
    cfg.StrOpt('auth_uri',
               default='http://controller:5000/v3.0',
               help=u._("Openstack identity url")),
    cfg.StrOpt('admin_service',
               default='service',
               help=u._("service project name")),
    cfg.StrOpt('admin_user',
               default='os_vm_expire',
               help=u._("os-vm-expire service user id")),
    cfg.StrOpt('admin_password',
               help=u._("os-vm-expire service user password")),
    cfg.StrOpt('admin_user_domain_name',
               default='default',
               help=u._("os-vm-expire user domain name")),
    cfg.StrOpt('admin_project_domain_name',
               default='default',
               help=u._("os-vm-expire service project domain name")),
]

queue_opt_group = cfg.OptGroup(name='queue',
                               title='Queue Application Options')

queue_opts = [
    cfg.BoolOpt('enable', default=False,
                help=u._('True enables queuing, False invokes '
                         'workers synchronously')),
    cfg.StrOpt('namespace', default='osvmexpire',
               help=u._('Queue namespace')),
    cfg.StrOpt('topic', default='osvmexpire.workers',
               help=u._('Queue topic name')),
    cfg.StrOpt('version', default='1.1',
               help=u._('Version of tasks invoked via queue')),
    cfg.StrOpt('server_name', default='osvmexpire.queue',
               help=u._('Server name for RPC task processing server')),
    cfg.IntOpt('asynchronous_workers', default=1,
               help=u._('Number of asynchronous worker processes')),
]


ks_queue_opt_group = cfg.OptGroup(name=KS_NOTIFICATIONS_GRP_NAME,
                                  title='Nova Notification Options')

ks_queue_opts = [
    cfg.BoolOpt('enable', default=False,
                help=u._('True enables nova notification listener '
                         ' functionality.')),
    cfg.StrOpt('control_exchange', default='nova',
               help=u._('The default exchange under which topics are scoped. '
                        'May be overridden by an exchange name specified in '
                        'the transport_url option.')),
    cfg.StrOpt('pool_name', default='os_vm_expire',
               help=u._('Pool notification to listen on nova exchange.'
                        'Messages in same pool will get messages distributed,'
                        ' while messages are copied over all pools')),
    cfg.StrOpt('topic', default='versioned_notifications',
               help=u._("nova notification queue topic name. This name "
                        "needs to match one of values mentioned in nova "
                        "deployment's 'notification_topics' configuration "
                        "e.g."
                        "    notification_topics=notifications.info, "
                        "    notifitions.error"
                        "Multiple servers may listen on a topic and messages "
                        "will be dispatched to one of the servers in a "
                        "round-robin fashion. That's why os-vm-expire service "
                        "should have its own dedicated notification queue so "
                        "that it receives all of nova notifications.")),
    cfg.BoolOpt('allow_requeue', default=False,
                help=u._('True enables requeue feature in case of notification'
                         ' processing error. Enable this only when underlying '
                         'transport supports this feature.')),
    cfg.StrOpt('version', default='1.0',
               help=u._('Version of tasks invoked via notifications')),
    cfg.IntOpt('thread_pool_size', default=10,
               help=u._('Define the number of max threads to be used for '
                        'notification server processing functionality.')),
]


context_opts = [
    cfg.StrOpt('admin_role', default='admin',
               help=u._('Role used to identify an authenticated user as '
                        'administrator.')),
    cfg.BoolOpt('allow_anonymous_access', default=False,
                help=u._('Allow unauthenticated users to access the API with '
                         'read-only privileges. This only applies when using '
                         'ContextMiddleware.')),
]

common_opts = [
    cfg.IntOpt('max_vm_duration',
               default=MAX_VM_DURATION_DAYS,
               help=u._("Maximum life duration of VM in days")),
    cfg.IntOpt('max_vm_extend',
               default=MAX_VM_EXTEND_DAYS,
               help=u._("Maximum life extend of VM in days")),
    cfg.IntOpt('max_vm_total_duration',
               default=MAX_VM_TOTAL_DURATION_DAYS,
               help=u._("Maximum life of VM in days, whatever the extends")),
]

host_opts = [
    cfg.StrOpt('host_href', default='http://localhost:9411',
               help=u._("Host name, for use in HATEOAS-style references Note: "
                        "Typically this would be the load balanced endpoint "
                        "that clients would use to communicate back with this "
                        "service. If a deployment wants to derive host from "
                        "wsgi request instead then make this blank. Blank is "
                        "needed to override default config value which is "
                        "'http://localhost:9411'")),
]


def list_opts():
    yield None, context_opts
    yield None, common_opts
    yield None, host_opts
    yield None, _options.eventlet_backdoor_opts
    yield queue_opt_group, queue_opts
    yield ks_queue_opt_group, ks_queue_opts
    yield cleaner_opt_group, cleaner_opts
    yield worker_opt_group, worker_opts
    yield mail_opt_group, mail_opts


# Flag to indicate  configuration is already parsed once or not
_CONFIG_PARSED_ONCE = False


def parse_args(conf, args=None, usage=None, default_config_files=None):
    global _CONFIG_PARSED_ONCE
    if default_config_files is None:
        dir_path = os.path.dirname(os.path.realpath(__file__))
        if 'OSVMEXPIRE_CONFIG' in os.environ and os.environ['OSVMEXPIRE_CONFIG']:
            default_config_files = [os.environ['OSVMEXPIRE_CONFIG']]
        elif os.path.exists('/etc/os-vm-expire/osvmexpire.conf'):
            default_config_files = ['/etc/os-vm-expire/osvmexpire.conf']
        elif os.path.exists('osvmexpire.conf'):
            default_config_files = [
                'osvmexpire.conf'
            ]
        elif os.path.exists(
            os.path.join(dir_path, '../../etc/os-vm-expire/osvmexpire.conf')
        ):
            default_config_files = [
                os.path.join(
                    dir_path,
                    '../../etc/os-vm-expire/osvmexpire.conf'
                )
            ]
        elif os.path.exists(os.path.join(dir_path, '../../osvmexpire.conf')):
            default_config_files = [
                os.path.join(dir_path, '../../osvmexpire.conf')
            ]

    conf(args=args if args else [],
         project='os-vm-expire',
         prog='os-vm-expire',
         version=os_vm_expire.version.__version__,
         usage=usage,
         default_config_files=default_config_files)

    conf.pydev_debug_host = os.environ.get('PYDEV_DEBUG_HOST')
    conf.pydev_debug_port = os.environ.get('PYDEV_DEBUG_PORT')

    # Assign cfg.CONF handle to parsed configuration once at startup
    # only. No need to keep re-assigning it with separate plugin conf usage
    if not _CONFIG_PARSED_ONCE:
        cfg.CONF = conf
        _CONFIG_PARSED_ONCE = True


def new_config():
    conf = cfg.ConfigOpts()
    log.register_options(conf)
    conf.register_opts(context_opts)
    conf.register_opts(common_opts)
    conf.register_opts(host_opts)
    conf.register_opts(_options.eventlet_backdoor_opts)
    conf.register_opts(_options.periodic_opts)

    conf.register_opts(_options.ssl_opts, "ssl")
    conf.register_group(queue_opt_group)
    conf.register_opts(queue_opts, group=queue_opt_group)
    conf.register_group(ks_queue_opt_group)
    conf.register_opts(ks_queue_opts, group=ks_queue_opt_group)
    conf.register_opts(cleaner_opts, group=cleaner_opt_group)
    conf.register_opts(worker_opts, group=worker_opt_group)
    conf.register_opts(mail_opts, group=mail_opt_group)

    # Update default values from libraries that carry their own oslo.config
    # initialization and configuration.
    set_middleware_defaults()

    return conf


def setup_remote_pydev_debug():
    """Required setup for remote debugging."""

    if CONF.pydev_debug_host and CONF.pydev_debug_port:
        try:
            try:
                from pydev import pydevd
            except ImportError:
                import pydevd

            pydevd.settrace(CONF.pydev_debug_host,
                            port=int(CONF.pydev_debug_port),
                            stdoutToServer=True,
                            stderrToServer=True)
        except Exception:
            LOG.exception('Unable to join debugger, please '
                          'make sure that the debugger processes is '
                          'listening on debug-host \'%(debug-host)s\' '
                          'debug-port \'%(debug-port)s\'.',
                          {'debug-host': CONF.pydev_debug_host,
                           'debug-port': CONF.pydev_debug_port})
            raise


def set_middleware_defaults():
    """Update default configuration options for oslo.middleware."""
    cors.set_defaults(
        allow_headers=['X-Auth-Token',
                       'X-Openstack-Request-Id',
                       'X-Project-Id',
                       'X-Identity-Status',
                       'X-User-Id',
                       'X-Storage-Token',
                       'X-Domain-Id',
                       'X-User-Domain-Id',
                       'X-Project-Domain-Id',
                       'X-Roles'],
        expose_headers=['X-Auth-Token',
                        'X-Openstack-Request-Id',
                        'X-Project-Id',
                        'X-Identity-Status',
                        'X-User-Id',
                        'X-Storage-Token',
                        'X-Domain-Id',
                        'X-User-Domain-Id',
                        'X-Project-Domain-Id',
                        'X-Roles'],
        allow_methods=['GET',
                       'PUT',
                       'POST',
                       'DELETE',
                       'PATCH']
    )


CONF = new_config()
LOG = logging.getLogger(__name__)
parse_args(CONF)

# Adding global scope dict for all different configs created in various
# modules.
_CONFIGS = {}


def set_module_config(name, module_conf):
    """Each plugin can set its own conf instance with its group name."""
    _CONFIGS[name] = module_conf


def get_module_config(name):
    """Get handle to plugin specific config instance by its group name."""
    return _CONFIGS[name]



# Copyright (c) 2013-2014 Rackspace, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from contextlib import contextmanager
import datetime
import functools
# import os
from os import path
import time
import types

import mock
from oslo_config import cfg
from oslo_utils import uuidutils
import oslotest.base as oslotest
from oslotest import createfile

import six
from six.moves.urllib import parse
import webtest

# from OpenSSL import crypto

from os_vm_expire.api import app
# from os_vm_expire.common import config
import os_vm_expire.context
# from os_vm_expire.model import repositories

from os_vm_expire.tests import database_utils


def mock_pecan_request(test_instance, host=None):
    patcher_obj = mock.patch('pecan.request')
    mock_req = patcher_obj.start()
    test_instance.addCleanup(patcher_obj.stop)
    mock_req.url = host


@contextmanager
def pecan_context(test_instance, host=None):
    mock_pecan_request(test_instance, host=host)
    yield


class OsVMExpireAPIBaseTestCase(oslotest.BaseTestCase):
    """Base TestCase for all tests needing to interact with a os-vm-expire app."""
    root_controller = None

    def _build_context(self, project_id, roles=None, user=None, is_admin=False,
                       policy_enforcer=None):
        context = os_vm_expire.context.RequestContext(
            roles=roles,
            user=user,
            project=project_id,
            is_admin=is_admin
        )
        context.policy_enforcer = policy_enforcer
        return context

    def setUp(self):
        super(OsVMExpireAPIBaseTestCase, self).setUp()
        # Make sure we have a test db and session to work with
        database_utils.setup_in_memory_db()

        # Generic project id to perform actions under
        self.project_id = generate_test_valid_uuid()

        # Build the test app
        wsgi_app = app.build_wsgi_app(
            controller=self.root_controller,
            transactional=True
        )

        self.app = webtest.TestApp(wsgi_app)
        self.app.extra_environ = {
            'os_vm_expire.context': self._build_context(self.project_id)
        }

    def tearDown(self):
        database_utils.in_memory_cleanup()
        super(OsVMExpireAPIBaseTestCase, self).tearDown()


class MockModelRepositoryMixin(object):
    """Class for setting up the repo factory mocks

    This class has the purpose of setting up the mocks for the model repository
    factory functions. This is because they are intended to be singletons, and
    thus called inside the code-base, and not really passed around as
    arguments. Thus, this kind of approach is needed.
    The functions assume that the class that inherits from this is a test case
    fixture class. This is because as a side-effect patcher objects will be
    added to the class, and also the cleanup of these patcher objects will be
    added to the tear-down of the respective classes.
    """

    def setup_vmexpire_repository_mock(self, mock_vmexpire_repo=mock.MagicMock(), ):
        """Mocks the expire repository factory function

        :param mock_expire_repo: The pre-configured mock repo to be
                                 returned.
        """
        self.mock_vmexpire_repo_patcher = None
        self._setup_repository_mock(repo_factory='get_vmexpire_repository',
                                    mock_repo_obj=mock_vmexpire_repo,
                                    patcher_obj=self.mock_vmexpire_repo_patcher)

    def setup_vmexclude_repository_mock(self, mock_vmexclude_repo=mock.MagicMock(), ):
        """Mocks the exclude repository factory function

        :param mock_exclude_repo: The pre-configured mock  repo to be
                                 returned.
        """
        self.mock_vmexclude_repo_patcher = None
        self._setup_repository_mock(repo_factory='get_vmexclude_repository',
                                    mock_repo_obj=mock_vmexclude_repo,
                                    patcher_obj=self.mock_vmexclude_repo_patcher)

    def _setup_repository_mock(self, repo_factory, mock_repo_obj, patcher_obj):
        patcher_obj = mock.patch(
            'os_vm_expire.model.repositories.' + repo_factory,
            return_value=mock_repo_obj
        )
        patcher_obj.start()
        self.addCleanup(patcher_obj.stop)


def construct_new_test_function(original_func, name, build_params):
    """Builds a new test function based on parameterized data.

    :param original_func: The original test function that is used as a template
    :param name: The fullname of the new test function
    :param build_params: A dictionary or list containing args or kwargs
        for the new test
    :return: A new function object
    """
    new_func = types.FunctionType(
        six.get_function_code(original_func),
        six.get_function_globals(original_func),
        name=name,
        argdefs=six.get_function_defaults(original_func),
        closure=six.get_function_closure(original_func)
    )

    for key, val in original_func.__dict__.items():
        if key != 'build_data':
            new_func.__dict__[key] = val

    # Support either an arg list or kwarg dict for our data
    build_args = build_params if isinstance(build_params, list) else []
    build_kwargs = build_params if isinstance(build_params, dict) else {}

    # Build a test wrapper to execute with our kwargs
    def test_wrapper(func, test_args, test_kwargs):
        @functools.wraps(func)
        def wrapper(self):
            return func(self, *test_args, **test_kwargs)
        return wrapper

    return test_wrapper(new_func, build_args, build_kwargs)


def process_parameterized_function(name, func_obj, build_data):
    """Build lists of functions to add and remove to a test case."""
    to_remove = []
    to_add = []

    for subtest_name, params in build_data.items():
        # Build new test function
        func_name = '{0}_{1}'.format(name, subtest_name)
        new_func = construct_new_test_function(func_obj, func_name, params)

        # Mark the new function as needed to be added to the class
        to_add.append((func_name, new_func))

        # Mark key for removal
        to_remove.append(name)

    return to_remove, to_add


def parameterized_test_case(cls):
    """Class decorator to process parameterized tests

    This allows for parameterization to be used for potentially any
    unittest compatible runner; including testr and py.test.
    """
    tests_to_remove = []
    tests_to_add = []
    for key, val in vars(cls).items():
        # Only process tests with build data on them
        if key.startswith('test_') and val.__dict__.get('build_data'):
            to_remove, to_add = process_parameterized_function(
                name=key,
                func_obj=val,
                build_data=val.__dict__.get('build_data')
            )
            tests_to_remove.extend(to_remove)
            tests_to_add.extend(to_add)

    # Add all new test functions
    [setattr(cls, name, func) for name, func in tests_to_add]

    # Remove all old test function templates (if they still exist)
    [delattr(cls, key) for key in tests_to_remove if hasattr(cls, key)]
    return cls


def parameterized_dataset(build_data):
    """Simple decorator to mark a test method for processing."""
    def decorator(func):
        func.__dict__['build_data'] = build_data
        return func
    return decorator


def setup_oslo_config_conf(testcase, content, conf_instance=None):

    conf_file_fixture = testcase.useFixture(
        createfile.CreateFileWithContent('osvmexpire', content))
    if conf_instance is None:
        conf_instance = cfg.CONF
    conf_instance([], project="os-vm-expire",
                  default_config_files=[conf_file_fixture.path])

    testcase.addCleanup(conf_instance.reset)


def create_timestamp_w_tz_and_offset(timezone=None, days=0, hours=0, minutes=0,
                                     seconds=0):
    """Creates a timestamp with a timezone and offset in days

    :param timezone: Timezone used in creation of timestamp
    :param days: The offset in days
    :param hours: The offset in hours
    :param minutes: The offset in minutes
    :return: a timestamp
    """
    if timezone is None:
        timezone = time.strftime("%z")

    timestamp = '{time}{timezone}'.format(
        time=(datetime.datetime.today() + datetime.timedelta(days=days,
                                                             hours=hours,
                                                             minutes=minutes,
                                                             seconds=seconds)),
        timezone=timezone)

    return timestamp


def get_limit_and_offset_from_ref(ref):
    matches = dict(parse.parse_qsl(parse.urlparse(ref).query))
    ref_limit = matches['limit']
    ref_offset = matches['offset']

    return ref_limit, ref_offset


def get_tomorrow_timestamp():
    tomorrow = (datetime.today() + datetime.timedelta(days=1))
    return tomorrow.isoformat()


def string_to_datetime(datetimestring, date_formats=None):
    date_formats = date_formats or [
        '%Y-%m-%d %H:%M:%S', '%Y-%m-%d %H:%M:%S.%f', '%Y-%m-%dT%H:%M:%S.%fZ',
        '%Y-%m-%dT%H:%M:%S.%f', "%Y-%m-%dT%H:%M:%SZ", "%Y-%m-%dT%H:%M:%S"]

    for dateformat in date_formats:
        try:
            return datetime.datetime.strptime(datetimestring, dateformat)
        except ValueError:
            continue
    else:
        raise


def get_id_from_ref(ref):
    """Returns id from reference."""
    ref_id = None
    if ref is not None and len(ref) > 0:
        ref_id = path.split(ref)[1]
    return ref_id


def generate_test_uuid(tail_value=0):
    """Returns a blank uuid with the given value added to the end segment."""
    return '00000000-0000-0000-0000-{value:0>{pad}}'.format(value=tail_value,
                                                            pad=12)


def generate_test_valid_uuid():
    """Returns a valid uuid value, similar to uuid generated in barbican"""
    return uuidutils.generate_uuid()


class DummyClassForTesting(object):
    pass
